#version 130

// "uniform" variables which are user-determined
uniform int stage;
uniform float proj33;
uniform float proj34;
uniform float depth_range;
uniform float current_depth;
uniform float filter_offset;
uniform float offset_scale;
uniform mat4 mvp_inv_mvp;
uniform sampler2D texred1;
uniform sampler2D texred2;

// "out" variables are the outputs of this pipeline stage to the next
// this just outputs a single float which will be written to red
out float fragment_colour;

// The main function runs once per "fragment" (per pixel)
void main () {
  vec2 viewsize, texcoord;
  float depth, this_red;

  // Always need texture coord for this fragment, and depth input
  viewsize = textureSize(texred2, 0);
  texcoord = gl_FragCoord.xy / viewsize;
  depth = texture(texred2, texcoord).r;

  // What we do depends on the stage:
  // 0 - nothing
  // 1 - orthographic or perspective depth linearisation
  // 2 - copy to (possibly lower resolution) output
  // 3 - horizontal filter
  // 4 - vertical filter (repeated several times in pair with horizontal)
  // 5 - convert to occlusion value

  if (stage == 1) {

    if (abs(proj34+1) > 0.0001) {
      // Convert perspective depth to linear depth in the red output component
      fragment_colour = proj34 / (proj33 - (depth * 2.0 - 1.0));
    } else {
      // Convert orthographic depth to linear depth in the red output component
      fragment_colour = ((depth * 2.0 - 1.0) - proj34) / proj33;
    }

  } else if (stage == 2) {

    // Just copy to output
    fragment_colour = depth;

  } else if (stage == 5) {

    // Ambient occlusion based on smoothed and original linear depth
    this_red = texture(texred1, texcoord).r;
    this_red = (this_red-depth)/depth_range;
    //fragment_colour = 6.0858*pow(clamp(this_red, 0.0, 0.3), 1.5);
    fragment_colour = clamp(this_red, 0.0, 1.0);

  } else {

    // Early exit if greater than current depth
    this_red = texture(texred1, texcoord).r;
    if (depth < current_depth) {
      fragment_colour = this_red;
    } else {
      float left_red, right_red, weight, left_depth, right_depth;
      vec2 offset;

      // filter based on input red values, either horizontal or vertical
      if (stage == 3) {
        offset = vec2(filter_offset, 0.0)/viewsize;
      } else {
        offset = vec2(0.0, filter_offset)/viewsize;
      }

      // Average centre with neighbouring samples
      left_red = texture(texred1, texcoord - offset).r;
      left_depth = texture(texred2, texcoord - offset).r;
      right_red = texture(texred1, texcoord + offset).r;
      right_depth = texture(texred2, texcoord + offset).r;
      // This ensures we don't allow objects more than depth_range in front
      // to affect the results
      //weight = clamp(2.0*(1.0 - (left_depth - depth)/depth_range), 0.0, 1.0);
      //weight *= clamp(2.0*(1.0 - (right_depth - depth)/depth_range), 0.0, 1.0);
      weight = smoothstep(0.0, 1.0, 2.0*(1.0 - (left_depth - depth)/depth_range));
      weight *= smoothstep(0.0, 1.0, 2.0*(1.0 - (right_depth - depth)/depth_range));
      // This ensures that we don't incorrectly shade surfaces at a tangent
      // to the viewing direction
      //weight *= clamp((2.0 - (this_red - left_red)/offset_scale), 0.0, 1.0);
      //weight *= clamp((2.0 - (this_red - right_red)/offset_scale), 0.0, 1.0);
      weight *= smoothstep(0.0, 1.0, 2.0 - (this_red - left_red)/offset_scale);
      weight *= smoothstep(0.0, 1.0, 2.0 - (this_red - right_red)/offset_scale);
      fragment_colour = (this_red + weight*(left_red + right_red))/(1.0 + 2.0*weight);
    }
  }
}
