#version 130

// "uniform" variables which are user-determined
uniform mat4 projection_matrix;
uniform vec4 lighting;
uniform vec4 light1_position;
uniform float radius;
uniform vec4 surface_diffuse;

// additional uniforms for if we are doing diffuse and specular shadows
uniform bool radiant_shadows;
uniform sampler2D texshadow;

// constants
const vec3 view = vec3(0.0, 0.0, 1.0);
const vec3 light2_direction = vec3(0.0, 0.0, 1.0);

// "in" variables are received from the previous pipeline stage
// "smooth" variables are bi-linearly interpolated between vertices
smooth in vec3 P;

// "out" variables are the outputs of this pipeline stage to the next
// "fragment" shaders should output an RGBA pixel value
out vec4 fragment_colour;

// The main function runs once per "fragment" (per pixel)
void main () {

  // Adjust amount of lighting
  vec3 light1_colour = vec3(lighting.r);
  float light2_colour = lighting.g;
  float ambient_colour = lighting.a;

  // Calculate shadowing if necessary
  if (radiant_shadows) {
    vec2 texcoord = gl_FragCoord.xy / textureSize(texshadow, 0);
    light1_colour *= texture(texshadow, texcoord).rgb;
  }

  // Find view point on background sphere with centre C
  float a, b, c, b2, ac4, depth;
  vec3 N, PC;
  vec3 C = vec3(0.0, 0.0, P.z + radius);

  // Different calculations for perspective and orthographic
  if (abs(projection_matrix[3][2]+1) > 0.0001) {

    // Solve for |Px - C|^2 = radius^2
    a = dot(P, P);
    //b = -2.0 * dot(P,C);
    b = -2.0 * P.z * C.z;
    //c = dot(C,C) - radius*radius;
    c = C.z * C.z - radius * radius;
    b2 = b * b;
    ac4 = 4.0 * a * c;
    if (ac4 > b2) {
      // outside this circle - no solutions
      depth = C.z;
      PC = vec3(P.xy, depth);
      N = vec3(0.0, 0.0, 1.0);
    } else {
      // inside this circle
      depth = P.z * (-b + sqrt(b2 - ac4)) / (2.0 * a);
      PC = vec3(P.xy, depth);
      N = normalize(C - PC);
    }

    // Turn depth back into ndc
    depth = (projection_matrix[2][2] * depth + projection_matrix[3][2]) / (projection_matrix[2][3] * depth);

  } else {

    // Solve for |P.x, P.y, depth|^2 = radius^2
    a = P.x*P.x + P.y*P.y;
    b = radius*radius;
    if (a > b) {
      // outside this circle - no solutions
      depth = C.z;
      PC = vec3(P.xy, depth);
      N = vec3(0.0, 0.0, 1.0);
    } else {
      // inside this circle
      depth = C.z - sqrt(b - a);
      PC = vec3(P.xy, depth);
      N = normalize(C - PC);
    }

    // Turn depth back into ndc
    depth = (projection_matrix[2][2] * depth + projection_matrix[3][2]);

  }

  // Set fragment depth - set this if also want to shadow against background
  //gl_FragDepth = (depth + 1.0) / 2.0;

  // Check for directional or point light 1
  vec3 light1_direction;
  if (light1_position.w != 0.0) {
    // In this case, the light1_position has to be already in view coordinates
    PC = light1_position.xyz - PC;
    float l1len = length(PC);
    float att = 1.0 / (1.0 + abs(light1_position.w) * pow(l1len, 2.0));
    light1_direction = PC / l1len;
    light1_colour *= att;
  } else {
    light1_direction = light1_position.xyz;
  }

  // Front facing by definition, but still need a colour if the light is behind
  float NdotL1 = 0.3*dot(N, light1_direction) + 0.7;
  if ((abs(light1_position.w) < 1e-10) && (light1_direction.z < 0.0)) NdotL1 = NdotL1 / (1.0 - light1_direction.z); // compensates for light increase
  float NdotL2 = max(dot(N, light2_direction), 0.0);

  // Output fragment colour
  fragment_colour = vec4(surface_diffuse.rgb * (light1_colour * NdotL1 + vec3(light2_colour * NdotL2 + ambient_colour)), surface_diffuse.a);
}
