#version 130

// #define VARYING_COL to include per-pixel diffuse colour

// "uniform" variables which are user-determined
uniform vec4 lighting;
uniform vec4 light1_position;
uniform bool radial_depth;
#ifndef VARYING_COL
uniform vec4 surface_diffuse;
#endif // VARYING_COL
uniform vec3 surface_material;

// additional uniforms for if we are doing depth peeling front-to-back translucency
uniform bool depth_peeling;
uniform sampler2D texdepth0;
uniform sampler2D texdepth1;

// additional uniforms for if we are doing diffuse and specular shadows
uniform bool radiant_shadows;
uniform sampler2D texshadow;

// additional uniforms for if we are doing ambient occlusion
uniform bool ambient_occlusion;
uniform sampler2D texambient;

// constants
const vec3 view = vec3(0.0, 0.0, 1.0);
const vec3 light2_direction = vec3(0.0, 0.0, 1.0);
const float alpha_intensity = 0.9;

// "in" variables are received from the previous pipeline stage
// "smooth" variables are bi-linearly interpolated between vertices
smooth in vec3 N;
smooth in vec3 L1;
smooth in vec3 P;
#ifdef VARYING_COL
smooth in vec4 surface_diffuse;
#endif // VARYING_COL

// "out" variables are the outputs of this pipeline stage to the next
// "fragment" shaders should output an RGBA pixel value
out vec4 fragment_colour;

// The main function runs once per "fragment" (per pixel)
void main () {

  // Check additional front and back depth maps and possibly discard
  if (depth_peeling) {
    vec2 texcoord = gl_FragCoord.xy / textureSize(texdepth0, 0);
    // Allow for worst case floating point precision between 0.0 and 1.0
    float depth0 = texture(texdepth0, texcoord).r - 0.0000001;
    float depth1 = texture(texdepth1, texcoord).r + 0.0000001;
    if ((gl_FragCoord.z >= depth0) || (gl_FragCoord.z <= depth1)) {
      fragment_colour =  vec4(0.0, 0.0, 0.0, 1.0);
      gl_FragDepth = depth0 - 0.0000001;
      return;
    }
  }
  if (radial_depth) {
    gl_FragDepth = 1.0 / gl_FragCoord.w;
  } else {
    gl_FragDepth = gl_FragCoord.z;
  }

  // Adjust amount of lighting
  vec3 light1_colour = vec3(lighting.r);
  vec3 light2_colour = vec3(lighting.g);
  vec3 surface_specular = vec3(lighting.b);
  vec3 ambient_colour = vec3(lighting.a);

  // Get material
  float surface_roughness = surface_material.r;
  float surface_metallic = surface_material.g;
  float surface_anisotropy = surface_material.b;

  // Calculate shadowing if necessary
  if (radiant_shadows) {
    vec2 texcoord = gl_FragCoord.xy / textureSize(texshadow, 0);
    light1_colour *= texture(texshadow, texcoord).rgb;
  }

  // colours mix between metal and non-metal material
  vec3 base = surface_diffuse.rgb;
  vec3 specular = mix(surface_specular, base, surface_metallic);
  vec3 diffuse = mix(base, 0.6*base, surface_metallic);

  // material parameters
  float rough = mix(0.2, 0.8, surface_roughness);
  float shininess = 2.0 / (rough * rough);
  specular *= 2.0*(1.0 - surface_roughness)*(1.0 - surface_roughness);

  // N must be re-normalised because it has been interpolated
  vec3 Nn = normalize(N);

  // Check for directional or point light 1
  vec3 light1_direction;
  if (light1_position.w != 0.0) {
    float l1len = length(L1);
    float att = 1.0 / (1.0 + abs(light1_position.w) * pow(l1len, 2.0));
    light1_direction = L1 / l1len;
    light1_colour *= att;
  } else {
    light1_direction = L1;
  }

  // Work out the alpha value from view position and normal
  //float alpha = abs(dot(Nn, normalize(P)));
  float alpha = abs(Nn.z);
  alpha = 1.0 - alpha_intensity * alpha;

  // Adjust ambient colour if doing ambient occlusion
  if (ambient_occlusion) {
    vec2 texcoord = gl_FragCoord.xy / textureSize(texambient, 0);
    ambient_colour *= (1.0 - 0.75 * alpha * texture(texambient, texcoord).r);
  }

  // Diffuse lighting based on alpha
  vec3 diffuse_lighting = (light1_colour + light2_colour) * alpha + ambient_colour;

  // Anisotropy
  vec3 tangent = cross(Nn, vec3(0.0, 1.0, 0.0));
  vec3 anis_normal = cross(tangent, view);
  anis_normal = cross(anis_normal, tangent);
  float anw = max(length(anis_normal), 0.0001);
  Nn = mix(Nn, anis_normal/anw, surface_anisotropy*smoothstep(0.0001, 0.0002, anw));

  // Specular lighting calculation
  vec3 R1 = -reflect(light1_direction, Nn);
  vec3 R2 = -reflect(light2_direction, Nn);
  float RdotV1 = pow(max(dot(R1, view), 0.0), shininess);
  float RdotV2 = pow(max(dot(R2, view), 0.0), shininess);
  vec3 specular_lighting = RdotV1 * light1_colour + RdotV2 * light2_colour;

  // Want specular highlights to be less opaque if front facing
  if (gl_FrontFacing) {
    alpha = mix(alpha, 0.5*(1.0+alpha), min(RdotV1 + RdotV2, 1.0-surface_roughness));
  }
  alpha = surface_diffuse.a * alpha;

  // Combine diffuse and specular lighting to get the final pixel value
  // may need to change alpha values for front-to-back translucency
  if (depth_peeling) {
    fragment_colour = vec4((diffuse * diffuse_lighting + specular * specular_lighting)*alpha, 1.0 - alpha);
  } else {
    fragment_colour = vec4(diffuse * diffuse_lighting + specular * specular_lighting, alpha);
  }
}
