#version 130

// #define VARYING_COL to include per-pixel diffuse colour

// "uniform" variables which are user-determined
uniform vec4 lighting;
uniform vec4 light1_position;
uniform bool radial_depth;
#ifndef VARYING_COL
uniform vec4 surface_diffuse;
#endif // VARYING_COL
uniform vec3 surface_material;

// additional uniforms for if we are doing depth peeling front-to-back translucency
uniform bool depth_peeling;
uniform sampler2D texdepth0;
uniform sampler2D texdepth1;

// additional uniforms for if we are doing diffuse and specular shadows
uniform bool radiant_shadows;
uniform sampler2D texshadow;

// additional uniforms for if we are doing ambient occlusion
uniform bool ambient_occlusion;
uniform sampler2D texambient;

// constants
const vec3 view = vec3(0.0, 0.0, 1.0);
const vec3 light2_direction = vec3(0.0, 0.0, 1.0);

// "in" variables are received from the previous pipeline stage
// "smooth" variables are bi-linearly interpolated between vertices
smooth in vec3 N;
smooth in vec3 L1;
#ifdef VARYING_COL
smooth in vec4 surface_diffuse;
#endif // VARYING_COL

// "out" variables are the outputs of this pipeline stage to the next
// "fragment" shaders should output an RGBA pixel value
out vec4 fragment_colour;
out vec4 fragment_data1;

             
void main() {

  // Check additional front and back depth maps and possibly discard
  if (depth_peeling) {
    vec2 texcoord = gl_FragCoord.xy / textureSize(texdepth0, 0);
    // Allow for worst case floating point precision between 0.0 and 1.0
    float depth0 = texture(texdepth0, texcoord).r - 0.0000001;
    float depth1 = texture(texdepth1, texcoord).r + 0.0000001;
    if ((gl_FragCoord.z >= depth0) || (gl_FragCoord.z <= depth1)) {
      fragment_colour =  vec4(0.0, 0.0, 0.0, 1.0);
      gl_FragDepth = depth0 - 0.0000001;
      return;
    }
  }
  if (radial_depth) {
    gl_FragDepth = 1.0 / gl_FragCoord.w;
  } else {
    gl_FragDepth = gl_FragCoord.z;
  }

  // Adjust amount of lighting
  vec3 light1_colour = vec3(lighting.r);
  vec3 light2_colour = vec3(lighting.g);
  vec3 surface_specular = vec3(lighting.b);
  vec3 ambient_colour = vec3(lighting.a);

  // Get material
  float surface_roughness = surface_material.r;
  float surface_metallic = surface_material.g;
  float surface_anisotropy = surface_material.b;

  // Calculate shadowing if necessary
  if (radiant_shadows) {
    vec2 texcoord = gl_FragCoord.xy / textureSize(texshadow, 0);
    light1_colour *= texture(texshadow, texcoord).rgb;
  }

  // Adjust ambient colour if doing ambient occlusion
  if (ambient_occlusion) {
    vec2 texcoord = gl_FragCoord.xy / textureSize(texambient, 0);
    ambient_colour *= (1.0 - 0.75 * surface_diffuse.a * texture(texambient, texcoord).r);
  }

  // colours mix between metal and non-metal material
  vec3 base = surface_diffuse.rgb;
  vec3 specular = mix(surface_specular, base, surface_metallic);
  vec3 diffuse = mix(base, 0.6*base, surface_metallic);

  // roughness with sensible range
  float rough = mix(0.2, 0.8, surface_roughness);
  float k = rough * rough;
  float ksq = k * k;

  // Check for directional or point light 1
  vec3 light1_direction;
  if (light1_position.w != 0.0) {
    float l1len = length(L1);
    float att = 1.0 / (1.0 + abs(light1_position.w) * pow(l1len, 2.0));
    light1_direction = L1 / l1len;
    light1_colour *= att;
  } else {
    light1_direction = L1;
  }

  // Diffuse dot products
  vec3 Nn = normalize(N);
  float NdotL1 = dot(Nn, light1_direction);
  float NdotL2 = dot(Nn, light2_direction);

  // Anisotropy
  vec3 tangent = cross(Nn, vec3(0.0, 1.0, 0.0));
  vec3 anis_normal = cross(tangent, view);
  anis_normal = cross(anis_normal, tangent);
  float anw = max(length(anis_normal), 0.0001);
  Nn = mix(Nn, anis_normal/anw, surface_anisotropy*smoothstep(0.0001, 0.0002, anw));

  // Specular dot products
  float NdotV = dot(Nn, view);
  vec3 H = normalize(light1_direction + view);
  float NdotH1 = max(dot(Nn, H), 0.001);
  float HdotV1 = max(dot(H, view), 0.001);
  H = normalize(light2_direction + view);
  float NdotH2 = max(dot(Nn, H), 0.001);
  float HdotV2 = max(dot(H, view), 0.001);
  if (gl_FrontFacing) {
    NdotV = max(NdotV, 0.001);
    NdotL1 = max(NdotL1, 0.001);
    NdotL2 = max(NdotL2, 0.001);
  } else {
    NdotV = max(-NdotV, 0.001);
    NdotL1 = max(-NdotL1, 0.001);
    NdotL2 = max(-NdotL2, 0.001);
  }

  // specular lighting
  vec3 fresnel = mix(specular, vec3(1.0), pow(1.0 - NdotV, 5.0)) * 0.12;

  // Distribution function
  float NdotHsq = NdotH1 * NdotH1;
  float d = NdotHsq * ksq;
  float D1 = exp((NdotHsq - 1.0) / d) / (NdotHsq * d);
  //float d = NdotHsq * ksq + (1.0 - NdotHsq);
  //float D1 = ksq / (d * d);
  NdotHsq = NdotH2 * NdotH2;
  d = NdotHsq * ksq;
  float D2 = exp((NdotHsq - 1.0) / d) / (NdotHsq * d);
  //float d = NdotHsq * ksq + (1.0 - NdotHsq);
  //float D2 = ksq / (d * d);

  // Geometry function
  float g = 2.0 * NdotH1 / HdotV1;
  float G1 = min(min(NdotV, NdotL1) * g, 1.0);
  g = 2.0 * NdotH2 / HdotV2;
  float G2 = min(min(NdotV, NdotL2) * g, 1.0);

  // Combined term
  vec3 specular1 = fresnel * G1 * D1 / (4.0 * NdotH1 * NdotV);
  vec3 specular2 = fresnel * G2 * D2 / (4.0 * NdotH2 * NdotV);

  // diffuse lighting
  diffuse *= (vec3(1.0) - 0.3*fresnel);

  // total radiant light
  light1_colour *= NdotL1;
  light2_colour *= NdotL2;
  vec3 total_light = (diffuse + specular1) * light1_colour + (diffuse + specular2) * light2_colour;
 
  // add in ambient lighting
  total_light += ambient_colour * base;

  // may need to change alpha values for front-to-back translucency
  if (depth_peeling) {
    fragment_colour = vec4(total_light*surface_diffuse.a, 1.0 - surface_diffuse.a);
  } else {
    fragment_colour = vec4(total_light, surface_diffuse.a);
  }
}