#version 130

// #define VARYING_COL to include per-pixel diffuse colour

// "uniform" variables which are user-determined
uniform mat4 modelview_matrix;
uniform mat4 modelviewproj_matrix;
uniform vec4 lighting;
uniform vec4 light1_position;
uniform bool radial_depth;
#ifndef VARYING_COL
uniform vec4 surface_diffuse;
#endif // VARYING_COL

// constants
const vec3 light2_direction = vec3(0.0, 0.0, 1.0);

// "in" variables are supplied from the vertex data
in vec4 position;
in vec3 normal;
#ifdef VARYING_COL
in vec4 colour;
#endif // VARYING_COL

// "out" variables are the outputs of this pipeline stage to the next
// "vertex" shaders should output the built-in variable gl_Position,
// as well as anything declared here
smooth out vec4 front_colour;
smooth out vec4 back_colour;

// The main function runs once per vertex
void main () {

  // Calculate the transformed normal vector for this vertex
  vec3 N = normalize(mat3(modelview_matrix) * normal);

  // Adjust amount of lighting
  vec3 light1_colour = vec3(lighting.r);
  vec3 light2_colour = vec3(lighting.g);
  vec3 ambient_colour = vec3(lighting.a);

  // Check for directional or point light 1
  vec3 light1_direction;
  if (light1_position.w != 0.0) {
    light1_direction = mat3(modelview_matrix) * (light1_position.xyz - position.xyz);
    float l1len = length(light1_direction);
    float att = 1.0 / (1.0 + abs(light1_position.w) * pow(l1len, 2.0));
    light1_direction /= l1len;
    light1_colour *= att;
  } else {
    light1_direction = normalize(light1_position.xyz);
  }

  // Diffuse lighting calculation
  float NdotL1 = dot(N, light1_direction);
  float NdotL2 = dot(N, light2_direction);
  vec3 front_diffuse = light1_colour * max(NdotL1, 0.0) + light2_colour * max(NdotL2, 0.0) + ambient_colour;
  vec3 back_diffuse = light1_colour * max(-NdotL1, 0.0) + light2_colour * max(-NdotL2, 0.0) + ambient_colour;

  // Total light calculation
#ifdef VARYING_COL
  front_colour = vec4(vec3(colour) * front_diffuse, colour[3]);
  back_colour = vec4(vec3(colour) * back_diffuse, colour[3]);
#else // VARYING_COL
  front_colour = vec4(vec3(surface_diffuse) * front_diffuse, surface_diffuse[3]);
  back_colour = vec4(vec3(surface_diffuse) * back_diffuse, surface_diffuse[3]);
#endif // VARYING_COL

  // Calculate the transformed and projected vertex position
  if (radial_depth) {
    vec4 ndc = modelviewproj_matrix * position;
    vec4 light1_ndc = modelviewproj_matrix * vec4(light1_position.xyz, 1.0);
    light1_ndc /= light1_ndc.w;
    float w, f;
    // get ndc and adjust for light location
    ndc.xyz /= ndc.w;
    ndc.xyz -= light1_ndc.xyz;
    if (light1_position.w < 0.0) ndc.z = -ndc.z;
    // project onto four planes which form a square-based pyramid from the light
    f = max(abs(ndc.x), abs(ndc.y));
    w = f + max(-f, ndc.z);
    w = max(w, 0.00001);
    // this changes z sign correctly so front-of-light is clipped
    // but need to use w (1/gl_Fragcoord.w) for depth in fragment shader
    ndc.z = (ndc.z - 1.0) * w;
    ndc.w = w;
    // scale x and y locations according to location of light
    ndc.xy += (w - f) * clamp(light1_ndc.xy, -1.0, 1.0);
    if (light1_position.w < 0.0) ndc.x = -ndc.x; // to preserve triangle winding
    ndc /= (max(abs(light1_ndc.x), abs(light1_ndc.y)) + abs(light1_ndc.z) + 2.0);
    gl_Position = ndc;
  } else {
    gl_Position = modelviewproj_matrix * position;
  }

}
