#version 130

// "uniform" variables which are user-determined
uniform int stage;
uniform float proj33;
uniform float proj34;
uniform mat4 mvp_inv_mvp;
uniform vec4 light1_ndc;
uniform bool radial_depth;
uniform vec2 depth_tolerance;
uniform sampler2D texdepth1;
uniform sampler2D texdepth2;
uniform sampler2D texdepth3;
uniform sampler2D texdepth4;
uniform sampler2D texdepth5;
uniform sampler2D texrgba1;
uniform sampler2D texrgba2;

// constant variables
mat3 w  = mat3(1.0/16.0, 1.0/8.0, 1.0/16.0, 1.0/8.0, 1.0/4.0, 1.0/8.0, 1.0/16.0, 1.0/8.0, 1.0/16.0);

// "out" variables are the outputs of this pipeline stage to the next
// this just outputs a single float which will be written to red
out vec3 fragment_colour;

// The main function runs once per "fragment" (per pixel)
void main () {
  vec2 viewsize, texcoord;
  float depth;

  // Always need texture coord for this fragment, and depth input
  viewsize = textureSize(texdepth1, 0);
  texcoord = gl_FragCoord.xy / viewsize;
  depth = texture(texdepth1, texcoord).r;

  // What we do depends on the stage:
  // 0 - nothing
  // 1 - just copy to output
  // 2 - shadow mapping
  // 3 - blurring - horizontal
  // 4 - blurring - vertical

  if (stage == 1) {

    // Just copy to output
    fragment_colour.r = depth;

  } else if (stage == 2) {

    // Calculate shadow texture
    float this_depth, shadow_o, shadow_t;
    int x, y;
    vec2 xyoff, xy, depthxy, shadow_depth;
    vec4 ndc;

    if (depth >= 1.0) {
      fragment_colour.rgb = vec3(1.0, 1.0, 1.0);
    } else { 

      // Get normalised device coordinates
      ndc.xyz = vec3(texcoord, depth) * 2.0 - 1.0;
      ndc.w = 1.0;

      // Undo perspective divide if necessary
      if (abs(proj34+1) > 0.0001) {
        ndc = ndc * (-proj34 / (proj33 - depth));
      }

      // Transform back through normal view and forward through shadow view
      ndc = mvp_inv_mvp * ndc;

      // Possibly adjust for point light
      bool front_buffer = false;
      if (radial_depth) {
        float w, f;
        // get ndc and adjust for light location
        ndc.xyz /= ndc.w;
        ndc.xyz -= light1_ndc.xyz;
        if (ndc.z < 0.0) {
          ndc.z = -ndc.z;
          front_buffer = true;
        }
        // project onto four planes which form a square-based pyramid from the light
        f = max(abs(ndc.x), abs(ndc.y));
        w = f + max(-f, ndc.z);
        w = max(w, 0.00001);
        float scale = max(abs(light1_ndc.x), abs(light1_ndc.y)) + abs(light1_ndc.z) + 2.0;
        ndc.z = w / scale; // as no_light.frag, but in that w = 1/w
        ndc.w = w;
        // scale x and y locations according to location of light
        ndc.xy += (w - f) * clamp(light1_ndc.xy, -1.0, 1.0);
        if (front_buffer) ndc.x = -ndc.x; // to preserve triangle winding
        ndc.z = (2.0 * ndc.z - 1.0) * ndc.w; // clip space transform
      }

      // Convert to texture coordinate
      ndc.xyz = 0.5 * (ndc.xyz / ndc.w + 1.0);
      xyoff = vec2(1.0, 1.0)/viewsize;

      // Compare with sampled opaque and translucent texture
      shadow_o = 0.0;
      shadow_t = 0.0;
      if ((ndc.z < 1.0) && (ndc.x > 0.0) && (ndc.x < 1.0) && (ndc.y > 0.0) && (ndc.y < 1.0)) { 
        for (x=0; x<=2; x++) {
          for (y=0; y<=2; y++) {
            xy = 0.5*vec2(x-1,y-1);
            xy *= xyoff;
            shadow_depth = ndc.z - depth_tolerance;
            if (front_buffer) {
              this_depth = texture(texdepth4, ndc.xy + xy).r;
            } else {
              this_depth = texture(texdepth2, ndc.xy + xy).r;
            }
            if (this_depth < shadow_depth.x) {
              shadow_o += w[x][y];
            }
            if (front_buffer) {
              this_depth = texture(texdepth5, ndc.xy + xy).r;
            } else {
              this_depth = texture(texdepth3, ndc.xy + xy).r;
            }
            if (this_depth < shadow_depth.y) {
              shadow_t += w[x][y];
            }
          }
        }
      }

      // Mix this with alpha and colour of that surface
      if (front_buffer) {
        ndc = texture(texrgba2, ndc.xy).rgba;
      } else {
        ndc = texture(texrgba1, ndc.xy).rgba;
      }
      float cr = max(shadow_o, shadow_t * ndc.a);
      float ct = (1.0 - cr);
      ct *= ct;
      ct *= ct;
      fragment_colour.rgb = (vec3(1.0, 1.0, 1.0) * ct + ndc.rgb * (1.0 - ct)) * (1.0 - cr);

    }

  } else {

    // Blur in horizonal or vertical direction
    float this_depth, this_w, w;
    vec3 this_shadow;
    vec2 offset;
    int i, n;

    if (stage == 3) {
      offset = vec2(1.0, 0.0)/viewsize;
    } else {
      offset = vec2(0.0, 1.0)/viewsize;
    }

    // The first element of mvp_inv_mvp doubles as the filter extent
    n = int(mvp_inv_mvp[0] + 0.5);
    w = 0.0;
    for (i=-n; i<=n; i++) {
      this_depth = texture(texdepth1, texcoord + offset*i).r;
      this_w = 1.0/(abs(depth - this_depth) + depth_tolerance.x);
      this_shadow += this_w * texture(texrgba1, texcoord + offset*i).rgb;
      w += this_w;
    }
    if (w > 0.0) {
      fragment_colour.rgb = this_shadow / w;
    } else {
      fragment_colour.rgb = vec3(0.0, 0.0, 0.0);
    }

  }
}
