#version 130

// #define RADIANT_SHADOWS to include shadowing from light1 and light2 (straight on)
// #define AMBIENT_SCATTER to also include scattering for ambient light
// #define OPAQUE_SHADOWS to also include shadowing on already drawn opaque surfaces
#define MAX_STEPS 24

// "uniform" variables which are user-determined
uniform int interpolation;
uniform float slice_skip;
uniform float slice_steps;
uniform vec4 lighting;
uniform vec4 light1_mv;
uniform float light1_z;
uniform vec3 light2_direction;
uniform vec3 gradient_x;
uniform vec3 gradient_y;
uniform vec3 gradient_z;
uniform vec3 pz;
uniform float gradient_threshold;
uniform float gradient_clarity;
uniform float wetness;
uniform float fwhm;
uniform bool no_shade_opaque;
uniform vec3 front_planes;
uniform vec3 back_planes;
uniform vec4 reslice_plane;
uniform vec3 landmark;
uniform vec3 landmark_radius;
uniform bool alpha_is_object;
uniform sampler3D texdata;
uniform sampler2D texdepth;
uniform sampler1D texmap1;
uniform sampler1D texmat1;
uniform sampler1D texmap2;
uniform sampler1D texmat2;

// additional uniforms for if we are doing diffuse shadows
uniform int occlusion;
#ifdef RADIANT_SHADOWS
uniform sampler2D texlight1;
uniform sampler2D texlight2;
#ifdef AMBIENT_SCATTER
uniform sampler2D texlight3;
#endif // AMBIENT_SCATTER
#ifdef OPAQUE_SHADOWS
uniform sampler2D texlight4;
#endif // OPAQUE_SHADOWS
#endif // RADIANT_SHADOWS

// "in" variables are received from the previous pipeline stage
// "smooth" variables are bi-linearly interpolated between vertices
smooth in vec3 L1;
smooth in vec3 vol_texcoord;
smooth in vec3 vol_position_z;
smooth in vec2 light2_clip;
#ifdef RADIANT_SHADOWS
smooth in vec3 light1_clip;
smooth in vec3 light1_clip_start;
#endif // RADIANT_SHADOWS

// "out" variables are the outputs of this pipeline stage to the next
// "fragment" shaders should output an RGBA pixel value
out vec4 fragment_colour;
#ifdef RADIANT_SHADOWS
out vec4 fragment_data1;
out vec4 fragment_data2;
#ifdef AMBIENT_SCATTER
out vec4 fragment_data3;
#endif // AMBIENT_SCATTER
#ifdef OPAQUE_SHADOWS
out vec4 fragment_data4;
#endif // OPAQUE_SHADOWS
#endif // RADIANT_SHADOWS


// standard random value generator
float rand (vec2 texcoord) {
  return fract(sin(dot(texcoord.xy ,vec2(12.9898,78.233))) * 43758.5453);
}


// Copyright (c) 2008-2009, Danny Ruijters. All rights reserved.
// http://www.dannyruijters.nl/cubicinterpolation/
// This file is part of CUDA Cubic B-Spline Interpolation (CI).
vec2 texture_cubic(sampler3D tex, vec3 coord) {

  // shift the coordinate from [0,1] to [-0.5, n-0.5]
  vec3 n = textureSize(tex, 0);
  vec3 icoord = coord * n - 0.5;
  vec3 i = floor(icoord);
  vec3 t = icoord - i;
  vec3 one_minus_t = 1.0 - t;

  vec3 w0 = 1.0/6.0 * one_minus_t * one_minus_t * one_minus_t;
  vec3 w1 = 2.0/3.0 - 0.5 * t * t * (2.0 - t);
  vec3 w2 = 2.0/3.0 - 0.5 * one_minus_t * one_minus_t * (2.0 - one_minus_t);
  vec3 w3 = 1.0/6.0 * t * t * t;

  vec3 g0 = w0 + w1;
  vec3 g1 = w2 + w3;
  vec3 h0 = ((w1 / g0) - 0.5 + i) / n;
  vec3 h1 = ((w3 / g1) + 1.5 + i) / n;

  // fetch the eight linear interpolations and weight
  vec2 tex000 = texture(tex, h0).rg;
  vec2 tex100 = texture(tex, vec3(h1.x, h0.y, h0.z)).rg;
  tex000 = mix(tex100, tex000, g0.x);
  vec2 tex010 = texture(tex, vec3(h0.x, h1.y, h0.z)).rg;
  vec2 tex110 = texture(tex, vec3(h1.x, h1.y, h0.z)).rg;
  tex010 = mix(tex110, tex010, g0.x);
  tex000 = mix(tex010, tex000, g0.y);
  vec2 tex001 = texture(tex, vec3(h0.x, h0.y, h1.z)).rg;
  vec2 tex101 = texture(tex, vec3(h1.x, h0.y, h1.z)).rg;
  tex001 = mix(tex101, tex001, g0.x);
  vec2 tex011 = texture(tex, vec3(h0.x, h1.y, h1.z)).rg;
  vec2 tex111 = texture(tex, h1).rg;
  tex011 = mix(tex111, tex011, g0.x);
  tex001 = mix(tex011, tex001, g0.y);

  return mix(tex001, tex000, g0.z);
}


// Return nearest neighbour value, required if this represents an object ID
float texture_nearest_g(sampler3D tex, vec3 coord)
{
  vec3 n = textureSize(tex, 0);
  //vec3 i = round(coord * n - 0.5);
  //return texture(tex, (i + 0.5)/n).g;
  return texelFetch(tex, ivec3(coord * n), 0).g;
}


// The main function runs once per "fragment" (per pixel)
void main () {

  // Output frame buffers:
  // fragment_colour - RGBA output (with blending in front-to-back and initially cleared to 0,0,0,1)
  // if RADIANT_SHADOWS is defined:
  // fragment_data1 - light1 output (with blending disabled and initially cleared to 1,1,1,1)
  // fragment_data2 - light2 output (with blending disabled and initially cleared to 1,1,1,1)
  // fragment_data3 - light3 (ambient) output (with blending disabled and initially cleared to a,a,a,1)
  // if OPAQUE_SHADOWS is defined:
  // fragment_data4 - shadowing on pre-rendered surfaces output (with blending disabled and initially cleared to 1,1,1,1)

  // Check if in front of depth texture, allowing a tolerance of the slice separation
  vec2 ts = textureSize(texdepth, 0);
  vec2 texcoord = gl_FragCoord.xy / ts;
  float depth = texture(texdepth, texcoord).r;
  float light2_tex_z = abs(0.5 * (light2_clip.x / light2_clip.y + 1.0) - gl_FragCoord.z);
  float this_depth = gl_FragCoord.z - 0.0*light2_tex_z; // Very fiddly to get this value right
  bool opaque = this_depth > depth;

  // Get light colours first
  float integer_light1_shift = 1.0;
  float lighting_g, lighting_r, lighting_ra;
  lighting_g = max(0.0, lighting.g); // a negative lighting.g is just used to compensate for large lighting.r in shadows

  // Check for directional or point light 1
  vec3 pn = normalize(pz);
  float att = 1.0;
  vec3 light1_direction = light1_mv.xyz;
  // Can't directly update lighting_r for attenuation, since the calculated value is the attenuation
  // from the point light location, not just between two different locations
  float l1len = 0.0;
  if (light1_mv.w > 0.0) {
    l1len = length(L1);
    att = 1.0 / (1.0 + light1_mv.w * pow(l1len, 2.0));
    light1_direction = L1 / l1len;
    integer_light1_shift = 0.8; // integer increments don't work as well for point lights
  }

#ifdef RADIANT_SHADOWS      
  vec3 light1_colour, light2_colour;
  vec4 light1, light2;

  // Convert light1 texture increments
  vec2 light1_tex_xy = 0.5 * (light1_clip.xy / light1_clip.z + 1.0) - texcoord;

  // Get light1
  if ((l1len > 0.0) && (((l1len*slice_skip) < (2.0*abs(pz.z))) || (dot(light1_direction, pn) < 0.1))) {
    // inject light close to point or sufficiently away from light transmission volume
    light1 = vec4(1.0);
  } else {
    light1 = texture(texlight1, texcoord + light1_tex_xy).rgba;
  }
  if (light1.a == 1.0) integer_light1_shift = 0.0; // No need if already full light
  light1_colour = mix(light1.rgb * light1.a, vec3(1.0), light1.a);

  // Get light2 and update given possible shadow spread for occlusion
  light2 = texture(texlight2, texcoord).rgba;
#ifndef AMBIENT_SCATTER
  // Only model ambient occlusion using light2 if we aren't doing ambient scatter using light3
  if (!opaque && (occlusion > 0.0) && (pz.z > 0.0)) {
    float occ = occlusion * 0.1 * abs(light2_tex_z);
    vec4 light2b = texture(texlight2, texcoord + vec2(occ, occ)).rgba;
    vec4 light2a = texture(texlight2, texcoord + vec2(-occ, occ)).rgba;
    vec4 light2c = texture(texlight2, texcoord + vec2(occ, -occ)).rgba;
    vec4 light2d = texture(texlight2, texcoord + vec2(-occ, -occ)).rgba;
    light2a = (light2a * (1.001 - light2a.a) + light2b * (1.001 - light2b.a) + light2c * (1.001 - light2c.a) + light2d * (1.001 - light2d.a)) / (4.004 - light2a.a - light2b.a - light2c.a - light2d.a);
    light2a = (light2 * (1.001 - light2.a) + light2a * (1.001 - light2a.a)) / (2.002 - light2.a - light2a.a);
#ifdef OPAQUE_SHADOWS
    // Attempt to not create apparently opaque objects just from spreading the front light
    light2 = mix(light2, light2a, smoothstep(0.02, 0.04, light2a.a));
#else // OPAQUE_SHADOWS
    light2 = light2a;
#endif // OPAQUE_SHADOWS
  }
#endif // AMBIENT_SCATTER
  light2_colour = mix(light2.rgb * light2.a, vec3(1.0), light2.a);

#ifdef AMBIENT_SCATTER
  // Get light3 (ambient scattering)
  vec4 light3 = vec4(0.0);
  if (occlusion > 0.0) {
    float scat = occlusion * 0.25 * abs(light2_tex_z);
    float i, j, nscat = max(1.0, round(scat*ts.x));
    float w, iw, nij = 0.0;
    vec2 vi, vj;
    if ((abs(pn.z) >= 0.99) || (length(light1_tex_xy) == 0.0)) {
      vi = vec2(scat/nscat);
      vj = vec2(-vi.y, vi.x);
      iw = 0.0;
    } else if (l1len > 0.0) {
      vj = normalize(light1_tex_xy) * 1.414214 * scat / nscat;
      vi = vec2(-vi.y, vi.x);
      vi *= (1.0 - length(pn.xy));
      iw = 0.5*length(pn.xy) / (nscat);
    } else {
      vi = normalize(light1_tex_xy) * 1.414214 * scat / nscat;
      vj = vec2(-vi.y, vi.x);
      vi *= (1.0 - 0.5*length(pn.xy));
      iw = 0.0;
    }
    for (i=-nscat; i<=nscat; i+=1.0) {
      w = (nscat - abs(i)) * iw + 1.0;
      for (j=-nscat; j<=nscat; j+=1.0) {
        if ((j*j + i*i) <= nscat*nscat) {
          light3 += w * texture(texlight3, texcoord + i*vi + j*vj).rgba;
          nij += w;
        }
      }
    }
    light3 /= nij;
  } else {
    light3 = texture(texlight3, texcoord).rgba;
  }
  light3.rgb *= light3.a;
  vec3 ambient_light, accum_ambient = vec3(0.0, 0.0, 0.0);
  float avg_shading = 0.0;
#endif // AMBIENT_SCATTER

#endif // RADIANT_SHADOWS

  // Calculate distance over which gradient is taken
  float gradient_distance, gradient_divisor;
  if (fwhm > 0.0) {
    gradient_distance = fwhm;
    gradient_divisor = min(gradient_distance, 1.0); // only scale for distances less than one pixel
  } else {
    gradient_distance = 1.0;
    gradient_divisor = -1.0; // Indicates not to correct data value
  }

  // Still want to add just a little random noise to account for number of steps
  // This has to be positive, so there is always one point just behind the edge
  float jitter = rand(texcoord) * 0.95/slice_steps;

  // We might repeat this whole process for multiple data samples to average out the response
  // initially according to slice_skip, but this might be adjusted later for large data gradients
  // **** Note - this moves *FORWARDS* from the current slice, even though the slices are processed
  // in *BACKWARDS* order **** (we can't run backwards, since there is no read access to alpha_dst)
  int nsteps = 1, n = 0;
  vec4 accum_lighting = vec4(0.0, 0.0, 0.0, 1.0);
  vec3 avg_colour = vec3(0.0, 0.0, 0.0);
  float light1_avg_alpha = 0.0;
  float light2_avg_alpha = 0.0;

  // Work out data texture coordinate vector and distance for movement to next slice at current screen coordinate
  vec3 tex_z = normalize(vol_position_z);
  float slice_distance = slice_skip / abs(dot(pn, tex_z));
  vec3 vol_texcoord_z = slice_distance * (gradient_x * tex_z.x + gradient_y * tex_z.y + gradient_z * tex_z.z);

  // For some reason, nsteps = 7 or 13 makes light1 and light2 turn off ... need to investigate!
  if (!opaque) nsteps = min(max(1, int(round(slice_steps))), MAX_STEPS);
  while (n < nsteps) {
    float frac = float(n)/float(nsteps);

    // Get current data point
    vec3 vol_texcoord_n = vol_texcoord - vol_texcoord_z * (frac + jitter);

    // Check first if we are outside the volume defined by all side planes
    float first_render = 2.0;
    float light1_shade = 1.0;
    float light2_shade = 1.0;
    vec2 txl, data1, data2;
    vec4 colour, colour_object;
    float data, sphere_alpha, data_alpha, data_object, data_fraction, alpha_a, alpha_b;
    float reslice_dist = (dot(vol_texcoord_n, reslice_plane.xyz) - reslice_plane.w);
    if (any(greaterThan(vec4((vol_texcoord_n - front_planes) * (vol_texcoord_n - back_planes), reslice_dist), vec4(0.0)))) {
      colour.a = 0.0;
      first_render = 0.0;
      integer_light1_shift = 0.0; // no need outside volume
#ifdef RADIANT_SHADOWS
      // any light copied from outside the volume should be full
      if (light1_mv.w != 0.0) light1 = vec4(1.0);
#endif // RADIANT_SHADOWS
    } else {
      // Also check for outside the spherical limit
      sphere_alpha = smoothstep(0.9, 1.1, length((vol_texcoord_n - landmark)/landmark_radius));
      if (landmark_radius.x >= 0.0) sphere_alpha = 1.0 - sphere_alpha;

      // Get linear data value here as a quick check for completely transparent voxel
      txl = texture(texdata, vol_texcoord_n).rg;
      data = txl.r;
      data_alpha = txl.g;
      colour = texture(texmap1, data).rgba;
      if (alpha_is_object) {
        data_object = texture_nearest_g(texdata, vol_texcoord_n);
        data_fraction = (data_object > 0.0) ? (data_alpha / data_object) : 0.0;
        //colour_object = texture(texmap2, (255.0*data_object + 0.5)/256.0);
        colour_object = texelFetch(texmap2, int(255.0*data_object + 0.5), 0);
        colour_object.a = pow(colour_object.a, 2.0);
        // Ensure that objects are always at least slightly visible, unless object alpha set lower
        if (colour_object.a > 0.5) {
          alpha_a = 0.02 * (2.0 * colour_object.a - 1.0);
          alpha_b = 1.0 - alpha_a;
        } else {
          alpha_a = 0.0;
          alpha_b = 2.0 * colour_object.a;
        }
        colour = mix(colour, vec4(colour_object.rgb, alpha_a + alpha_b * colour.a), data_fraction);
        data_alpha = 0.0;
      }
      colour.a = colour.a * sphere_alpha;
    }

    // Very little to do if this data is completely transparent
    if ((colour.a > 0.0) && (data_alpha < 1.0)) {
      // Update opacity for this specific location
      this_depth = gl_FragCoord.z - light2_tex_z * (frac + jitter);

      // If using point light, check light direction is consistent with slice direction
#ifdef RADIANT_SHADOWS
      this_depth = (this_depth - light1_z)*sign(pz.z)/abs(light2_tex_z);
      lighting_ra = lighting.r * clamp(0.05*(this_depth + 12.0), 0.0, 1.0); // switch-over at this_depth = -2.0, width 20
      lighting_r = lighting.r * clamp(0.25*(this_depth + 2.0), 0.0, 1.0); // switch-over at this_depth = 0.0, width 4
#else // RADIANT_SHADOWS
      lighting_r = lighting.r;
#endif // RADIANT_SHADOWS

      // Might need to access data again if doing cubic interpolation
      if (interpolation == 2) {
        txl = texture_cubic(texdata, vol_texcoord_n).rg;
        data = txl.r;
        colour = texture(texmap1, data).rgba;
        data_alpha = txl.g;
        if (alpha_is_object) {
          colour = mix(colour, vec4(colour_object.rgb, alpha_a + alpha_b * colour.a), data_fraction);
          data_alpha = 0.0;
        }
        colour.a = colour.a * sphere_alpha;
      }

      // Might need to adjust shading if we are right at the front edge of the rendered volume
      vec3 pf;
      vec4 vdist, adist;
      float fraction;
      vdist.xyz = (vol_texcoord_n - front_planes) / vol_texcoord_z;
      vdist.w = reslice_dist / dot(vol_texcoord_z, reslice_plane.xyz);
      vdist *= slice_distance; // compensates for vol_texcoord changing length for differently aligned planes pn
      adist = abs(vdist);
      fraction = min(min(min(adist.x, adist.y), adist.z), adist.w);
      if (fraction < 2.0) {
        // Work out which plane, and check it is actually facing forward
        if (fraction == adist.w) {
          fraction = vdist.w;
          pf = vec3(dot(reslice_plane.xyz, gradient_x), dot(reslice_plane.xyz, gradient_y), dot(reslice_plane.xyz, gradient_z));
        } else if (fraction == adist.x) {
          fraction = vdist.x;
          pf = vec3(gradient_x.x, gradient_y.x, gradient_z.x);
        } else if (fraction == adist.y) {
          fraction = vdist.y;
          pf = vec3(gradient_x.y, gradient_y.y, gradient_z.y);
        } else {
          fraction = vdist.z;
          pf = vec3(gradient_x.z, gradient_y.z, gradient_z.z);
        }
        if (fraction >= 0.0) {
          // within slice distance of front plane, so render constant according to plane normal
          first_render = fraction;
          if (gradient_threshold < 1.0) {
            pf = normalize(pf) / sign(pf.z);
            // end planes have basic diffuse lighting only
#ifdef RADIANT_SHADOWS
            // Looks better if we only light one side, due to shadowing
            light1_shade = max(dot(pf, light1_direction), 0.0);
            light2_shade = max(dot(pf, light2_direction), 0.0);
#else // RADIANT_SHADOWS
            // Looks better if we light both sides, since the data gradient is not signed
            light1_shade = abs(dot(pf, light1_direction));
            light2_shade = abs(dot(pf, light2_direction));
#endif // RADIANT_SHADOWS
            // Want to also apply gradient clarity to the end planes
            if (gradient_clarity > 0.0) {
              colour.a *= pow(1.0 - light2_shade, gradient_clarity);
            }
          }
        }
      }

      if (fraction < 4.0) {
        // Need consistent light direction close to end planes otherwise can see the integer
        // pattern over the surface
        integer_light1_shift = 0.0;
      }

      // Whether to do shading of this pixel, always disabled for edge data with no gradient value
      float shading, shade_at_all, shade;
      if (opaque) {
        shading = 0.0; // This can slightly affect the transmission of light1, but not by enough to matter
      } else {
        shading = smoothstep(1.0, 2.0, first_render);
      }
      shade_at_all = shading;

      // Get material, and diffuse and ambient light properties
      vec3 mat = texture(texmat1, data).rgb;
      if (alpha_is_object) {
        //vec3 data_mat = texture(texmat2, data_object).rgb;
        vec3 data_mat = texelFetch(texmat2, int(255.0*data_object + 0.5), 0).rgb;
        mat = mix(mat, data_mat, data_fraction);
      }
      float surface_metallic = float(mat.g);
      vec3 diffuse_colour = mix(1.0, 0.6, surface_metallic) * colour.rgb;

      // Possibly get data gradient in modelview direction
      vec3 lighting1, lighting2, ambient1, ambient2;
      bool some_light = true;
      // For front-to-back, only need to do this loop if the front light is still non-zero (i.e. can still see this fragment)
      // For back-to-front, need to keep going if any light is non-zero
#ifdef RADIANT_SHADOWS
#ifdef AMBIENT_SCATTER
      some_light = (light2.a >= 0.01) || ((pz.z < 0.0) && ((light3.a >= 0.01) || (light1.a*lighting_r >= 0.0001)));
#else // AMBIENT_SCATTER
      some_light = (light2.a >= 0.01) || ((pz.z < 0.0) && (light1.a*lighting_r >= 0.0001));
#endif // AMBIENT_SCATTER
#endif // RADIANT_SHADOWS
      if ((shading > 0.0) && (colour.a > 0.0) && some_light) {

        float f1, f2;
        vec3 p, pd;
        if (interpolation == 2) {

          // cubic gradient
          p.x = texture_cubic(texdata, vol_texcoord_n + 0.1 * gradient_x).r - data;
          p.y = texture_cubic(texdata, vol_texcoord_n + 0.1 * gradient_y).r - data;
          p.z = texture_cubic(texdata, vol_texcoord_n + 0.1 * gradient_z).r - data;
          p = p / max(length(p), 0.00000001);

          // sample data at +/- gradient points
          pd = p * gradient_distance;
          data1 = texture_cubic(texdata, vol_texcoord_n + pd.x * gradient_x + pd.y * gradient_y + pd.z * gradient_z).rg;
          data2 = texture_cubic(texdata, vol_texcoord_n - pd.x * gradient_x - pd.y * gradient_y - pd.z * gradient_z).rg;

        } else {
            
          // simple, linear gradient
          p.x = texture(texdata, vol_texcoord_n + gradient_x).r;
          p.y = texture(texdata, vol_texcoord_n + gradient_y).r;
          p.z = texture(texdata, vol_texcoord_n + gradient_z).r;
          p.x = p.x - texture(texdata, vol_texcoord_n - gradient_x).r;
          p.y = p.y - texture(texdata, vol_texcoord_n - gradient_y).r;
          p.z = p.z - texture(texdata, vol_texcoord_n - gradient_z).r;
          p = p / max(length(p), 0.0001);
             
          // sample data at +/- gradient points
          pd = p * gradient_distance;
          data1 = texture(texdata, vol_texcoord_n + pd.x * gradient_x + pd.y * gradient_y + pd.z * gradient_z).rg;
          data2 = texture(texdata, vol_texcoord_n - pd.x * gradient_x - pd.y * gradient_y - pd.z * gradient_z).rg;

        }

        // Ensure gradient points towards direction of reduced opacity
        //f1 = texture(texmap1, data1.r).a;
        //f2 = texture(texmap1, data2.r).a;
        //if (f1 < f2) p = -p;
        // Ensure gradient always pointing forwards
        if (p.z < 0.0) p = -p;

        // Work out if we need more steps and adjust shading, given the data gradient
        f1 = (data1.r - data) / abs(gradient_divisor);
        f2 = (data - data2.r) / abs(gradient_divisor);
        shade = max(abs(f1), abs(f2)) * 2.0;
        if (n==0) {
          nsteps = min(max(int(round(shade * 12.0)), nsteps), MAX_STEPS);
        }
        if (gradient_threshold == 1.0) {
          shade = 0.0;
        } else {
          shade = smoothstep(0.0, gradient_threshold, shade);
        }
        shading *= shade;

        // Reduce shading if asked to for opaque material
        if (no_shade_opaque) {
          shading *= smoothstep(0.0, 0.05, 1.0 - colour.a);
        }

        // Get specular and diffuse light properties only if we are still shading
        if (shading > 0.0) {

          // Handle transitions which are due to blurring
          float wet_rough = 1.0;
          float trans = 1.0;
          if (gradient_divisor > 0.0) {
            trans = min(abs(f1), abs(f2)) * (sign(f1 * f2) + 1.0);
            if ((fraction > 2.0) && (trans > 0.0) && !alpha_is_object) {
              // This option changes the data value to one of the surrounding regions
              trans = smoothstep(0.0, gradient_threshold, trans) * pow(abs(p.z), 0.1);
              if (abs(f1) > abs(f2)) {
                txl = mix(txl, data2, trans);
              } else {
                txl = mix(txl, data1, trans);
              }
              data = txl.r;
              data_alpha = txl.g;
              trans = colour.a;
              colour = texture(texmap1, data);
              colour.a *= sphere_alpha;
              colour.a = min(colour.a, trans); // this softens the edge a tiny bit so it isn't so pixelated
              mat = texture(texmat1, data).rgb;
              surface_metallic = float(mat.g);
              diffuse_colour = mix(1.0, 0.6, surface_metallic) * colour.rgb;
              wet_rough = mix(1.0, 1.0 - trans, 0.75 * wetness);
              trans = 1.0;
            } else {
              // This version keeps the data value but makes it transparent
              trans = smoothstep(gradient_threshold, 2.0 * gradient_threshold, trans);
              trans = 1.0 - trans * pow(abs(p.z), 0.25);
              wet_rough = mix(1.0, trans, wetness);
            }
          }

          // Highlight outlines by reducing opacity and dependence on light direction
          if (gradient_clarity > 0.0) {
            colour.a *= pow(1.0 - mix(1.0, abs(dot(p, light2_direction)), shade), gradient_clarity);
            shading *= max((1.0 - 0.1 * gradient_clarity), 0.0);
          }
          
          // Diffuse lighting calculation
          float ap1 = dot(p, light1_direction);
          integer_light1_shift *= smoothstep(0.1, 0.3, abs(ap1)); // reduce for grazing light
#ifdef RADIANT_SHADOWS
          // Just light front side
          lighting1 = mix(light1_shade, max(ap1, 0.0), shading) * diffuse_colour;
          lighting2 = mix(light2_shade, max(dot(p, light2_direction), 0.0), shading) * diffuse_colour;
#ifdef AMBIENT_SCATTER
          float scatter = mix(0.5, 0.5*pow(1.0 - p.z*p.z, 0.5), shading);
          ambient1 = lighting1 * scatter;
          ambient2 = lighting2 * scatter;
          avg_shading += shading;
#endif // AMBIENT_SCATTER
#else // RADIANT_SHADOWS
          // Light both sides
          lighting1 = mix(light1_shade, abs(ap1), shading) * diffuse_colour;
          lighting2 = mix(light2_shade, abs(dot(p, light2_direction)), shading) * diffuse_colour;
#endif // RADIANT_SHADOWS

          // Specular lighting calculation
          // For wet surfaces, specular light still shines through transition regions
          float surface_roughness = float(mat.r);
          surface_roughness *= wet_rough;
          vec3 specular_colour = mix(vec3(mix(2.0, lighting.b, wet_rough)), colour.rgb, surface_metallic);
          float rough = mix(0.2, 0.8, surface_roughness);
          float shininess = 2.0 / (rough * rough);
          specular_colour *= 2.0*(1.0 - surface_roughness)*(1.0 - surface_roughness);

          // Adjust p for anisotropy
          float surface_anisotropy = float(mat.b);
          if (surface_anisotropy > 0.0) {
            vec3 tangent = cross(p, vec3(0.0, 1.0, 0.0));
            vec3 anis_normal = cross(tangent, vec3(0.0, 0.0, 1.0));
            anis_normal = cross(anis_normal, tangent);
            float anw = max(length(anis_normal), 0.0001);
            p = mix(p, anis_normal/anw, surface_anisotropy*smoothstep(0.0001, 0.0002, anw));
          }

          // Add in specular lighting
          float spec1, spec2;
          spec1 = mix(0.0, pow(max(-reflect(light1_direction, p).z, 0.0), shininess), shading);
          spec2 = mix(0.0, pow(max(-reflect(light2_direction, p).z, 0.0), shininess), shading);
          lighting1 += spec1 * specular_colour;
          lighting2 += spec2 * specular_colour;
          // Still see specular light in transition regions for wet surfaces
          trans = mix(trans, 1.0, wetness * min(spec1*lighting_r + spec2*lighting_g, 1.0 - surface_roughness));
          colour.a *= trans;

        } else {

          // No shading based on small gradient
          lighting1 = light1_shade * diffuse_colour;
          lighting2 = light2_shade * diffuse_colour;
#ifdef AMBIENT_SCATTER
          ambient1 = lighting1 * 0.5;
          ambient2 = lighting2 * 0.5;
#endif // AMBIENT_SCATTER

          // But still need to reduce opacity if the gradient was poorly defined
          if ((gradient_threshold < 1.0) && (gradient_clarity > 0.0)) colour.a *= shade;

        }

      } else {

        // Lighting not affected by data gradient, or no light
        lighting1 = light1_shade * diffuse_colour;
        lighting2 = light2_shade * diffuse_colour;
#ifdef AMBIENT_SCATTER
        ambient1 = lighting1 * 0.5;
        ambient2 = lighting2 * 0.5;
#endif // AMBIENT_SCATTER
        if (first_render < 1.0) {
          shade = 1.0;
        } else {
          shade = 0.0;
        }

      }

      // Take account of point light attenuation in the lighting of data, but not in the stored light transfer
      lighting1 *= att;
#ifdef AMBIENT_SCATTER
      // Only apply ambient scattering model if we aren't lighting a front plane (which shows interpolation artefacts)
      vec3 lighting_ambient = diffuse_colour * mix(vec3(lighting.a), light3.rgb, shade_at_all);
      ambient1 *= att;
#else // AMBIENT_SCATTER
      vec3 lighting_ambient = diffuse_colour * lighting.a;
#endif // AMBIENT_SCATTER

      // The second component of the data texture might be a translucency value
      // If this does not exist, then it is set to zero (opaque) by default
      colour.a *= (1.0 - data_alpha);
     
      // Data may have become transparent due to previous operations
      if (colour.a > 0.0) {

        // Adjust accum alpha for multiple steps and also for slice_distance
        float accum_alpha2 = (1.0 - pow(1.0 - colour.a, slice_distance));
        float accum_alpha = (1.0 - pow(1.0 - accum_alpha2, 1.0/float(nsteps)));

        // Adjust avg_alpha for light_distance
        float light1_distance = slice_skip / max(abs(dot(pn, light1_direction)), 0.1);
        //float avg_alpha = accum_alpha; // this is best for shading gradient lines
        //float avg_alpha = accum_alpha2; // this is best for lighting consistency
        // So we mix between them: with the former taking precedence if gradient was higher
        float light1_alpha2 = (1.0 - pow(1.0 - colour.a, light1_distance));
        float light1_alpha = (1.0 - pow(1.0 - light1_alpha2, 1.0/float(nsteps)));
        light1_alpha = mix(light1_alpha2, light1_alpha, shade);
        float light2_alpha = mix(accum_alpha2, accum_alpha, shade);

        // Accumulate average and summed lighting for each step
        avg_colour += colour.rgb;
        light1_avg_alpha += (1.0 - light1_alpha);
        light2_avg_alpha += (1.0 - light2_alpha);

        // light_colour kept constant so accum_lighting calculated now, which is not
        // completely correct, but faster
        if (!opaque) {
          // Apply the light colour to the lighting calculation
#ifdef RADIANT_SHADOWS
          if (first_render < 1.0) {
#endif // RADIANT_SHADOWS
            lighting1 *= lighting_r;
            lighting2 *= lighting_g;
#ifdef RADIANT_SHADOWS
#ifdef AMBIENT_SCATTER
            ambient1 *= lighting_ra; // continue calculation even if lighting_r is zero
            ambient2 *= lighting_g;
#endif // AMBIENT_SCATTER
          } else {
            lighting1 *= light1_colour * lighting_r;
            lighting2 *= light2_colour * lighting_g;
#ifdef AMBIENT_SCATTER
            ambient1 *= light1_colour * lighting_ra; // continue calculation even if lighting_r is zero
            ambient2 *= light2_colour * lighting_g;
#endif // AMBIENT_SCATTER
          }
#ifdef AMBIENT_SCATTER
          accum_ambient = accum_ambient * (1.0 - accum_alpha) + (ambient1 + ambient2) * accum_alpha;
#endif // AMBIENT_SCATTER
#endif // RADIANT_SHADOWS
          accum_lighting.rgb = accum_lighting.rgb * (1.0 - accum_alpha) + (lighting1 + lighting2 + lighting_ambient) * accum_alpha;
          accum_lighting.a *= (1.0 - accum_alpha);
        } else {
#ifdef RADIANT_SHADOWS
#ifndef OPAQUE_SHADOWS
#ifndef AMBIENT_SCATTER
          // This fixes occlusion causing things to appear opaque when they aren't and generating spaces around objects
          // But it also makes the shadowing onto opaque surfaces go wrong (i.e. puts them in permanent shadow)
          // Not needed if we are doing ambient scattering rather than ambient occlusion
          accum_lighting.a = 0.0;
#endif // AMBIENT_SCATTER
#endif // OPAQUE_SHADOWS
#endif // RADIANT_SHADOWS
        }

      } else {

        // Data is transparent
        avg_colour += colour.rgb;
        light1_avg_alpha += 1.0;
        light2_avg_alpha += 1.0;
        
      }

    } else {

      // Data is transparent
      avg_colour += colour.rgb;
      light1_avg_alpha += 1.0;
      light2_avg_alpha += 1.0;

    }

    n += 1;
  } // end of loop for steps n

  // Update the main RGBA output
  fragment_colour = accum_lighting;

#ifdef RADIANT_SHADOWS
#ifdef OPAQUE_SHADOWS
  // For front-to-back, keep updating the shadow while the data is in front of the opaque surfaces
  // For back-to-front, keep updating the shadow while the data is behind the opaque surfaces
  // Only do any of this if the directional light is enabled, otherwise just write the input to output only
  vec4 light4;
  if (depth > 0.995) {
    // No shadows over the background
    light4 = vec4(1.0, 1.0, 1.0, 0.0);
  } else {
    bool test;
    if (pz.z > 0.0) {
      test = (gl_FragCoord.z + 0.0*light2_tex_z) >= depth;
    } else {
      test = (gl_FragCoord.z + 1.0*light2_tex_z) <= depth;
    }
    if (test) {
      // Just write the existing light back to the output
      light4 = texture(texlight4, texcoord).rgba;
      if (light4.a == 1.0) {
        // Might need to blur in light1 direction if increment is more than one pixel 
        if ((lighting.r > 0.0) && ((L1.z * pz.z) >= 0.0)) {
          float li, ilen;
          ilen = length(light1_tex_xy * ts);
          for (li = 1.0; li < ilen; li += 1.0) {
            light4 += texture(texlight4, texcoord + light1_tex_xy * li/ilen).rgba;
          }
          light4.rgb /= li;
        }
        light4.a = 0.0; // So don't blur again later whatever
      }
    } else {
      // New shadow calculation
      float light1_alpha;
      this_depth = (gl_FragCoord.z - light1_z)*sign(pz.z)/abs(light2_tex_z) + 0.0;
      if (this_depth < 1.0) {
        // This allows a bit of overlap in shading if going from two directions. It is complicated here by the fact
        // that the shading is multiplied rather than added in outline_canvas (which is quicker)
        light1_alpha = 1.0 - (1.0 - pow(1.0 - (1.0 - light1.a)*lighting.r, max((this_depth + 1.0)/2.0, 0.0)))/lighting.r;
      } else {
        light1_alpha = light1.a;
      }
      vec3 colour1 = mix(light1.rgb * light1.a, vec3(1.0), light1_alpha) * lighting.r * att;
      vec3 colour2 = mix(light2.rgb * light2.a, vec3(1.0), light2.a) * lighting_g;
      // lighting.r may need compensating for directional light - see setup_lighting() in outline_canvas.cpp
      if (abs(light1_mv.w) < 1e-10) {
        lighting_r = lighting.r / max(1.0, 1.0 - light1_direction.z);
      } else {
        lighting_r = lighting.r;
      }
      // Note that lighting.g can be negative to compensate for a lighting_r which has been set larger
#ifdef AMBIENT_SCATTER
      light4 = vec4(vec3(1.0 - lighting_r - lighting.g - lighting.a) + colour1 + colour2 + light3.rgb, 1.0);
#else // AMBIENT_SCATTER
      light4 = vec4(vec3(1.0 - lighting_r - lighting.g) + colour1 + colour2, 1.0);
#endif // AMBIENT_SCATTER
    }
  }

  // Update lighting for opaque shadows
  fragment_data4 = light4;
#endif // OPAQUE_SHADOWS

  // Get average colour and apply to lighting update
  avg_colour /= float(nsteps);
  light1_avg_alpha /= float(nsteps);
  light2_avg_alpha /= float(nsteps);
  // Using light1_tex_xy directly introduces significant (fairly random) blur due to multiple linear interpolations
  // Using round() with tsf = ts removes blur completely by always using whole-texel shifts
  // Adding a small factor to ts re-introduces blur, but controlled and much more uniform over slice spacing
  // However, turn this off when close to end planes, or for grazing light over a surface, since then
  // you can see the integer light pattern
  if (integer_light1_shift > 0.0) {
    vec2 light1_tex_start = 0.5 * (light1_clip_start.xy / light1_clip_start.z + 1.0) - texcoord;
    vec2 tsf = ts + 0.04 * occlusion / ((vec2(1.0) / ts) + length(light1_tex_xy));
    vec2 new_light1_tex_xy = ( round(light1_tex_start * tsf) - round((light1_tex_start - light1_tex_xy) * tsf)) / tsf;
    light1 = mix(light1, texture(texlight1, texcoord + new_light1_tex_xy).rgba, integer_light1_shift);
  }
  if (light1_avg_alpha < 1.0) {
    //vec3 ct = mix(avg_colour.rgb, light1.rgb, pow(light1_avg_alpha, 4.0));
    vec3 ct = light1.rgb * pow(avg_colour.rgb, vec3(1.0 - light1_avg_alpha));
    light1 = vec4(ct, light1.a * light1_avg_alpha);
  }
  if (light2_avg_alpha < 1.0) {
    //vec3 ct = mix(avg_colour.rgb, light2.rgb, pow(light2_avg_alpha, 4.0));
    vec3 ct = light2.rgb * pow(avg_colour.rgb, vec3(1.0 - light2_avg_alpha));
    light2 = vec4(ct, light2.a * light2_avg_alpha);
  }

  // Update main lighting buffers
  fragment_data1 = light1;
  fragment_data2 = light2;

#ifdef AMBIENT_SCATTER
  // Update scattered light
  avg_shading = 1.0 - avg_shading / float(nsteps);
  accum_ambient /= float(nsteps);
  fragment_data3 = vec4(light3.rgb + accum_ambient, 1.0 - (1.0 - light2_avg_alpha) * avg_shading);// * (1.0 - accum_lighting.a), light2_avg_alpha);
#endif // AMBIENT_SCATTER

  // Calculate what to write to the depth buffer - can only update depth for front-to-back
  this_depth = gl_FragCoord.z + 1.0*light2_tex_z; // Very fiddly to get this value right
  if ((pz.z > 0.0) && (light2.a < 0.01) && (this_depth < depth)) {
    // Update the opaque depth buffer to give depth indication at the end of volume rendering
    gl_FragDepth = this_depth;
  } else {
    // No change - just write out current value to flip-flopped depth buffer
    gl_FragDepth = depth;
  }
#endif // RADIANT_SHADOWS

}

